// DirectoryTreeCtrl.cpp : implementation file
//
/////////////////////////////////////////////
// written by robert rostek - tecxx@rrs.at //
/////////////////////////////////////////////

#include "stdafx.h"
#include "DirectoryTreeCtrl.h"
#include "otherfunctions.h"
#include "emule.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

// CDirectoryTreeCtrl

IMPLEMENT_DYNAMIC(CDirectoryTreeCtrl, CTreeCtrl)
CDirectoryTreeCtrl::CDirectoryTreeCtrl()
{
	m_bSelectSubDirs = false;
	m_bCtrlPressed = false;
}

CDirectoryTreeCtrl::~CDirectoryTreeCtrl()
{
//	Don't destroy the system's image list
	m_image.Detach();
}

BEGIN_MESSAGE_MAP(CDirectoryTreeCtrl, CTreeCtrl)
	ON_NOTIFY_REFLECT(TVN_ITEMEXPANDING, OnTvnItemexpanding)
	ON_NOTIFY_REFLECT(TVN_GETDISPINFO, OnTvnGetdispinfo)
	ON_WM_LBUTTONDOWN()
	ON_NOTIFY_REFLECT(NM_RCLICK, OnNMRclickSharedList)
	ON_NOTIFY_REFLECT(TVN_KEYDOWN, OnTvnKeydown)
	ON_WM_KEYDOWN()
	ON_WM_KEYUP()
END_MESSAGE_MAP()



// CDirectoryTreeCtrl message handlers

void CDirectoryTreeCtrl::OnTvnItemexpanding(NMHDR *pNMHDR, LRESULT *pResult)
{
	LPNMTREEVIEW pNMTreeView = reinterpret_cast<LPNMTREEVIEW>(pNMHDR);
	HTREEITEM hItem = pNMTreeView->itemNew.hItem;
	HTREEITEM hRemove = GetChildItem(hItem);

//	remove all subitems
	while(hRemove)
	{
		DeleteItem(hRemove);
		hRemove = GetChildItem(hItem);
	}

//	get the directory
	CString strDir = GetFullPath(hItem);

//	fetch all subdirectories and add them to the node
	AddSubdirectories(hItem, strDir);

	*pResult = 0;
}

void CDirectoryTreeCtrl::OnLButtonDown(UINT nFlags, CPoint point)
{
	UINT uFlags;
	HTREEITEM hItem = HitTest(point, &uFlags);
	HTREEITEM tItem = GetFirstVisibleItem();

	if((hItem) && (uFlags & TVHT_ONITEMSTATEICON))
	{
		CheckChanged(hItem, !GetCheck(hItem));

		if(nFlags & MK_CONTROL)
		{
			Expand(hItem, TVE_TOGGLE);
			HTREEITEM hChild;
			hChild = GetChildItem(hItem);
			while (hChild != NULL)
			{
				MarkChilds(hChild,!GetCheck(hItem));
				hChild = GetNextSiblingItem(hChild);
			}
			Expand(hItem, TVE_TOGGLE);
		}
	}
	SelectSetFirstVisible(tItem);
	CTreeCtrl::OnLButtonDown(nFlags, point);
}

void CDirectoryTreeCtrl::MarkChilds(HTREEITEM hChild, bool mark)
{
	CheckChanged(hChild, mark);
	SetCheck(hChild,mark);
	Expand(hChild, TVE_TOGGLE);
	HTREEITEM hChild2;
	hChild2 = GetChildItem(hChild);

	while( hChild2 != NULL)
	{
		MarkChilds(hChild2,mark);
		hChild2 = GetNextSiblingItem( hChild2 );
	}

	Expand(hChild, TVE_TOGGLE);
}

void CDirectoryTreeCtrl::OnTvnGetdispinfo(NMHDR *pNMHDR, LRESULT *pResult)
{
	LPNMTVDISPINFO pTVDispInfo = reinterpret_cast<LPNMTVDISPINFO>(pNMHDR);
	pTVDispInfo->item.cChildren = 1;
	*pResult = 0;
}

void CDirectoryTreeCtrl::Init(void)
{
	DeleteAllItems();

	WORD wWinVer = g_eMuleApp.m_pGlobPrefs->GetWindowsVersion(); // maybe causes problems on 98 & nt4
	if(wWinVer == _WINVER_2K_ || wWinVer == _WINVER_XP_ || wWinVer == _WINVER_ME_ || wWinVer == _WINVER_SE_)		
	{
		SHFILEINFO shFinfo;
		HIMAGELIST hImgList = NULL;

	//	Get the system image list using a "path" which is available on all systems. [patch by bluecow]
		hImgList = (HIMAGELIST)SHGetFileInfo(_T("."), 0, &shFinfo, sizeof(shFinfo), SHGFI_SYSICONINDEX | SHGFI_SMALLICON);
		m_image.m_hImageList = hImgList;
		SetImageList(&m_image, TVSIL_NORMAL);
	}

	TCHAR drivebuffer[500];
	::GetLogicalDriveStrings(500, drivebuffer); // e.g. "a:\ c:\ d:\"
	const TCHAR* pos = drivebuffer;
	while(*pos != '\0')
	{
	//	Copy drive name
		TCHAR drive[4];
		memccpy(drive, pos, '\0', 4);
		switch(drive[0])
		{
			case 'a':
			case 'A':
			case 'b':
			case 'B':
			//	Skip floppy disk
				break;
			default:
				drive[2] = '\0';
				AddChildItem(NULL, drive); // e.g. ("c:")
		}
	//	Point to the next drive (4 chars interval)
		pos = &pos[4];
	}
}

HTREEITEM CDirectoryTreeCtrl::AddChildItem(HTREEITEM hRoot, CString strText)
{
	CString strPath = GetFullPath(hRoot);

	if (hRoot != NULL && strPath.Right(1) != "\\")
		strPath += "\\";

	CString strDir = strPath + strText;
	TV_INSERTSTRUCT itInsert;

	memzero(&itInsert, sizeof(itInsert));
	
	WORD wWinVer = g_eMuleApp.m_pGlobPrefs->GetWindowsVersion();

	if(wWinVer == _WINVER_2K_ || wWinVer == _WINVER_XP_ || wWinVer == _WINVER_ME_ || wWinVer == _WINVER_SE_)
	{
		itInsert.item.mask = TVIF_CHILDREN | TVIF_HANDLE | TVIF_TEXT | TVIF_STATE | TVIF_IMAGE | TVIF_SELECTEDIMAGE;
		itInsert.item.stateMask = TVIS_BOLD | TVIS_STATEIMAGEMASK;
	}
	else
	{
		itInsert.item.mask = TVIF_CHILDREN | TVIF_HANDLE | TVIF_TEXT | TVIF_STATE;
		itInsert.item.stateMask = TVIS_BOLD;
	}
	
	if (HasSharedSubdirectory(strDir))
		itInsert.item.state = TVIS_BOLD;
	else
		itInsert.item.state = 0;

	if (HasSubdirectories(strDir))
		itInsert.item.cChildren = I_CHILDRENCALLBACK;		// used to display the + symbol next to each item
	else
		itInsert.item.cChildren = 0;

	itInsert.item.pszText = strText.GetBuffer();
	itInsert.item.cchTextMax = strText.GetLength();
	itInsert.hInsertAfter = TVI_SORT;
	itInsert.hParent = hRoot;
	
	if(wWinVer == _WINVER_2K_ || wWinVer == _WINVER_XP_ || wWinVer ==  _WINVER_ME_|| wWinVer == _WINVER_SE_)		
	{
		CString strTemp = strDir;

		if(strTemp.Right(1) != "\\")
			strTemp += "\\";
		
		UINT nType = GetDriveType(strTemp);

		if(DRIVE_REMOVABLE <= nType && nType <= DRIVE_RAMDISK)
			itInsert.item.iImage = nType;

		SHFILEINFO shFinfo;

		if(!SHGetFileInfo(strTemp, 0, &shFinfo,	sizeof(shFinfo), SHGFI_ICON | SHGFI_SMALLICON))
		{
			itInsert.itemex.iImage = 0;
		}
		else
		{
			itInsert.itemex.iImage = shFinfo.iIcon;
			DestroyIcon(shFinfo.hIcon);
		}
		
		if(!SHGetFileInfo(strTemp, 0, &shFinfo, sizeof(shFinfo), SHGFI_ICON | SHGFI_OPENICON | SHGFI_SMALLICON))
		{
			itInsert.itemex.iImage = 0;
		}
		else
		{
			itInsert.itemex.iSelectedImage = shFinfo.iIcon;
			DestroyIcon(shFinfo.hIcon);
		}
	}
	
	HTREEITEM hItem = InsertItem(&itInsert);

	if (IsShared(strDir))
		SetCheck(hItem);

	strText.ReleaseBuffer();

	return hItem;
}

CString CDirectoryTreeCtrl::GetFullPath(HTREEITEM hItem)
{
	CString strDir;
	HTREEITEM hSearchItem = hItem;

	while(hSearchItem != NULL)
	{
		strDir = GetItemText(hSearchItem) + _T('\\') + strDir;
		hSearchItem = GetParentItem(hSearchItem);
	}

	return strDir;
}

void CDirectoryTreeCtrl::AddSubdirectories(HTREEITEM hRoot, CString strDir)
{
	if (strDir.Right(1) != '\\')
		strDir += '\\';

	if (!::SetCurrentDirectory(strDir))
		return;

	CFileFind finder;
	BOOL bWorking = finder.FindFile(_T("*.*"));

	while (bWorking)
	{
		bWorking = finder.FindNextFile();

		if (finder.IsDots())
			continue;
		if (finder.IsSystem())
			continue;
		if (!finder.IsDirectory())
			continue;
		
		CString strFilename = finder.GetFileName();

		if (strFilename.ReverseFind('\\') != -1)
			strFilename = strFilename.Mid(strFilename.ReverseFind('\\') + 1);

		AddChildItem(hRoot, strFilename);
	}

	finder.Close();
}

bool CDirectoryTreeCtrl::HasSubdirectories(CString strDir)
{
	if (strDir.Right(1) != '\\')
		strDir += '\\';

	::SetCurrentDirectory(strDir);

	CFileFind finder;
	BOOL bWorking = finder.FindFile(_T("*.*"));

	while (bWorking)
	{
		bWorking = finder.FindNextFile();

		if (finder.IsDots())
			continue;
		if (finder.IsSystem())
			continue;
		if (!finder.IsDirectory())
			continue;

		finder.Close();

		return true;
	}

	finder.Close();
	return false;
}


void CDirectoryTreeCtrl::GetSharedDirectories(CStringList* list)
{
	for (POSITION pos = m_lstShared.GetHeadPosition(); pos != NULL;)
		list->AddTail(m_lstShared.GetNext(pos));
}

bool DirectoryExists(const CString Name)
{
	DWORD Code = GetFileAttributes(Name);
	return (Code != -1) && (Code & FILE_ATTRIBUTE_DIRECTORY);
}


void CDirectoryTreeCtrl::SetSharedDirectories(CStringList* list)
{
	m_lstShared.RemoveAll();

	for (POSITION pos = list->GetHeadPosition(); pos != NULL;)
	{
		CString str = list->GetNext(pos);

		if (str.Right(1) != '\\')
			str += '\\';

		m_lstShared.AddTail(str);
	}

	Init();
}

bool CDirectoryTreeCtrl::HasSharedSubdirectory(CString strDir)
{
	if (strDir.Right(1) != _T('\\'))
		strDir += _T('\\');

	strDir.MakeLower();

	for (POSITION pos = m_lstShared.GetHeadPosition(); pos != NULL;)
	{
		CString str = m_lstShared.GetNext(pos);

		str.MakeLower();

		if (str.Find(strDir) == 0 && strDir != str) //strDir.GetLength() != str.GetLength())
			return DirectoryExists(str);
	}

	return false;
}

void CDirectoryTreeCtrl::CheckChanged(HTREEITEM hItem, bool bChecked)
{
	CString strDir = GetFullPath(hItem);

	if (bChecked)
		AddShare(strDir);
	else
		DelShare(strDir);

	UpdateParentItems(hItem);

	CWnd *pParent = GetParent();

	if(pParent)
		pParent->SendMessage(WM_COMMAND, DIRLIST_ITEMSTATECHANGED, (long)m_hWnd);
}

bool CDirectoryTreeCtrl::IsShared(CString strDir)
{
	if (strDir.Right(1) != _T('\\'))
		strDir += _T('\\');

	for (POSITION pos = m_lstShared.GetHeadPosition(); pos != NULL;)
	{
		CString str = m_lstShared.GetNext(pos);

		if (str.Right(1) != _T('\\'))
			str += _T('\\');

		if (str.CompareNoCase(strDir) == 0)
			return true;
	}

	return false;
}

void CDirectoryTreeCtrl::AddShare(CString strDir)
{
	if (strDir.Right(1) != ('\\'))
		strDir += _T('\\');
	
	if (IsShared(strDir) || !strDir.CompareNoCase(CString(g_eMuleApp.m_pGlobPrefs->GetConfigDir())))
		return;

	m_lstShared.AddTail(strDir);
}

void CDirectoryTreeCtrl::DelShare(CString strDir)
{
	if (strDir.Right(1) != _T('\\'))
		strDir += _T('\\');

	for (POSITION pos = m_lstShared.GetHeadPosition(); pos != NULL;)
	{
		POSITION pos2 = pos;
		CString str = m_lstShared.GetNext(pos);

		if (str.CompareNoCase(strDir) == 0)
			m_lstShared.RemoveAt(pos2);
	}
}

void CDirectoryTreeCtrl::UpdateParentItems(HTREEITEM hChild)
{
	HTREEITEM hSearch = GetParentItem(hChild);

	while(hSearch != NULL)
	{
		if (HasSharedSubdirectory(GetFullPath(hSearch)))
			SetItemState(hSearch, TVIS_BOLD, TVIS_BOLD);
		else
			SetItemState(hSearch, 0, TVIS_BOLD);

		hSearch = GetParentItem(hSearch);
	}
}

void CDirectoryTreeCtrl::OnNMRclickSharedList(NMHDR *pNMHDR, LRESULT *pResult)
{
//	Get item under cursor
	POINT point;
	::GetCursorPos(&point);
	CPoint p = point;
	ScreenToClient(&p);
	HTREEITEM hItem = HitTest(p);

	CTitleMenu  menuShared;

//	Create the menu
	menuShared.CreatePopupMenu();
	menuShared.AddMenuTitle(GetResString((m_lstShared.GetCount() == 0) ? IDS_NOSHAREDFOLDERS : IDS_SHAREDFOLDERS));

	bool		bFolder = false;

//	Add right clicked folder, if any
	if (hItem)
	{
		m_strLastRightClicked = GetFullPath(hItem);

		if (!IsShared(m_strLastRightClicked))
		{
			menuShared.AppendMenu(MF_STRING, MP_SHAREDFOLDERS_FIRST-1, (LPCTSTR)(GetResString(IDS_VIEW1) + m_strLastRightClicked +GetResString(IDS_VIEW2)));
			bFolder = true;
		}
	}

//	Add all shared directories
	int iCnt = 0;

	for (POSITION pos = m_lstShared.GetHeadPosition(); pos != NULL; iCnt++)
		menuShared.AppendMenu(MF_STRING,MP_SHAREDFOLDERS_FIRST+iCnt, (LPCTSTR)m_lstShared.GetNext(pos));

//	Display menu (do not display empty menu)
	if ((m_lstShared.GetHeadPosition() != NULL) || bFolder)
	{
		menuShared.TrackPopupMenuEx(TPM_LEFTALIGN | TPM_RIGHTBUTTON, point.x, point.y, this, NULL);
	}
	*pResult = 0;
//	Menu objects are destroyed in their destructor
}

BOOL CDirectoryTreeCtrl::OnCommand(WPARAM wParam,LPARAM lParam )
{
	if (wParam < MP_SHAREDFOLDERS_FIRST)
	{
		ShellExecute(NULL, _T("open"), m_strLastRightClicked, NULL, NULL, SW_SHOW);
		return false;
	}

	int cnt = 0;

	for (POSITION pos = m_lstShared.GetHeadPosition(); pos != NULL;)
	{
		CString str = m_lstShared.GetNext(pos);

		if (cnt == wParam-MP_SHAREDFOLDERS_FIRST)
		{
			ShellExecute(NULL, _T("open"), str, NULL, NULL, SW_SHOW);
			return true;
		}

		cnt++;
	}

	return true;
}

void CDirectoryTreeCtrl::OnTvnKeydown(NMHDR *pNMHDR, LRESULT *pResult)
{ 
	LPNMTVKEYDOWN pTVKeyDown = reinterpret_cast<LPNMTVKEYDOWN>(pNMHDR);

	switch(pTVKeyDown->wVKey)
	{
		case VK_SPACE:
		case VK_RETURN:
		{
			HTREEITEM hItem = GetSelectedItem();

			if(hItem)
			{		
				HTREEITEM tItem = GetFirstVisibleItem();

				CheckChanged(hItem, !GetCheck(hItem));
								
				if(m_bCtrlPressed)
				{
					Expand(hItem, TVE_TOGGLE);
					HTREEITEM hChild;
					hChild = GetChildItem(hItem);

					while(hChild != NULL)
					{ 
						MarkChilds(hChild,!GetCheck(hItem));
						hChild = GetNextSiblingItem( hChild );
					}

					SetCheck(hItem, !GetCheck(hItem));
					Expand(hItem, TVE_TOGGLE);
				}
		 
				SelectSetFirstVisible(tItem);
			}

			break;
		}

		default:
			break;
	}

	*pResult = 0;
}

void CDirectoryTreeCtrl::OnKeyDown(UINT nChar, UINT nRepCnt, UINT nFlags)
{
	if(nChar == VK_CONTROL)
		m_bCtrlPressed = true;

	CTreeCtrl::OnKeyDown(nChar, nRepCnt, nFlags);
}

void CDirectoryTreeCtrl::OnKeyUp(UINT nChar, UINT nRepCnt, UINT nFlags)
{
	if(nChar == VK_CONTROL)
		m_bCtrlPressed = false;

	CTreeCtrl::OnKeyUp(nChar, nRepCnt, nFlags);
}
