//This program is free software; you can redistribute it and/or
//modify it under the terms of the GNU General Public License
//as published by the Free Software Foundation; either
//version 2 of the License, or (at your option) any later version.
//
//This program is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.
//
//You should have received a copy of the GNU General Public License
//along with this program; if not, write to the Free Software
//Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

#include "stdafx.h"
#include "IrcSocket.h"
#include "packets.h"
#include "opcodes.h"
#include "ircmain.h"
#include "emule.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif


CIrcSocket::CIrcSocket(CIrcMain* pIrcMain) : CAsyncSocketEx()
{
	m_pIrcMain = pIrcMain;
	m_pProxyLayer = NULL;
}

CIrcSocket::~CIrcSocket()
{
	RemoveAllLayers();
}

BOOL CIrcSocket::Create(UINT nSocketPort, int nSocketType, long lEvent, LPCTSTR lpszSocketAddress)
{
	EMULE_TRY

	ProxySettings settings = g_eMuleApp.m_pGlobPrefs->GetProxy();
	if (settings.m_bUseProxy && settings.m_nType!=PROXYTYPE_NOPROXY)
	{
		m_pProxyLayer=new CAsyncProxySocketLayer;
		switch (settings.m_nType)
		{
			case PROXYTYPE_SOCKS4:
				m_pProxyLayer->SetProxy(PROXYTYPE_SOCKS4,settings.m_strName,settings.m_uPort);
				break;
			case PROXYTYPE_SOCKS4A:
				m_pProxyLayer->SetProxy(PROXYTYPE_SOCKS4A,settings.m_strName,settings.m_uPort);
				break;
			case PROXYTYPE_SOCKS5:
				if (settings.m_bEnablePassword)
					m_pProxyLayer->SetProxy(PROXYTYPE_SOCKS5,settings.m_strName, settings.m_uPort, settings.m_strUser, Decrypt(settings.m_strPassword));
				else
					m_pProxyLayer->SetProxy(PROXYTYPE_SOCKS5,settings.m_strName,settings.m_uPort);
				break;
			case PROXYTYPE_HTTP11:
				if (settings.m_bEnablePassword)
					m_pProxyLayer->SetProxy(PROXYTYPE_HTTP11,settings.m_strName,settings.m_uPort, settings.m_strUser, Decrypt(settings.m_strPassword));
				else
					m_pProxyLayer->SetProxy(PROXYTYPE_HTTP11,settings.m_strName,settings.m_uPort);
				break;
			default:
 				ASSERT(0);
		}
		AddLayer(m_pProxyLayer);
	}

	return CAsyncSocketEx::Create(nSocketPort, nSocketType, lEvent, lpszSocketAddress);

	EMULE_CATCH

	return false;
}

void CIrcSocket::Connect()
{
	CAsyncSocketEx::Connect(g_eMuleApp.m_pGlobPrefs->GetIRCServer(), 6667);
}

void CIrcSocket::OnReceive(int nErrorCode) 
{
	if (nErrorCode){
		AddLogLine(false, RGB_LOG_ERROR + _T("IRC socket: Failed to read"));
		return;
	}

	TCHAR buffer[256];
	int length = Receive(buffer, sizeof(buffer)-1);	
	if (length < 0){
		AddLogLine(false, RGB_LOG_ERROR + _T("IRC socket: Failed to read"));
		return;
	}
	if (length > 0)
	{
		buffer[length] = '\0';
                g_eMuleApp.m_pDownloadQueue->AddDownDataOverheadOther(length*8);
		m_pIrcMain->PreParseMessage( buffer );
	}
}

void CIrcSocket::OnConnect(int nErrorCode)
{
	if (nErrorCode){
		AddLogLine(true, RGB_LOG_ERROR + _T("IRC socket: Failed to connect"));
		return;
	}
	m_pIrcMain->SetConnectStatus(true);
	m_pIrcMain->SendLogin();
}
void CIrcSocket::OnClose(int nErrorCode)
{
	if (nErrorCode){
		AddLogLine(false, RGB_LOG_ERROR + _T("IRC socket: Failed to close"));
		return;
	}
	RemoveAllLayers();
	m_pIrcMain->SetConnectStatus(false);
	CAsyncSocketEx::Close();
}

int CIrcSocket::SendString(CString message){
	message += _T("\r\n");
	int size = message.GetLength();
	g_eMuleApp.m_pUploadQueue->AddUpDataOverheadOther(size);
	return Send(message, size);
}

void CIrcSocket::RemoveAllLayers()
{
	EMULE_TRY
	CAsyncSocketEx::RemoveAllLayers();
	
	if (m_pProxyLayer) 
	{
		delete m_pProxyLayer;
		m_pProxyLayer = NULL;
	}
	EMULE_CATCH
}

int CIrcSocket::OnLayerCallback(const CAsyncSocketExLayer *pLayer, int nType, int nParam1, int nParam2)
{
	EMULE_TRY
	if (nType == LAYERCALLBACK_LAYERSPECIFIC)
	{
		ASSERT( pLayer );
		if (pLayer==m_pProxyLayer)
		{
			switch (nParam1)
			{
			case PROXYERROR_NOCONN:
				AddLogLine(false, RGB_LOG_ERROR + GetResString(IDS_ERRORMSG_PROXY_NOCONN));
				break;
			case PROXYERROR_REQUESTFAILED:
				AddLogLine(false, RGB_LOG_ERROR + GetResString(IDS_ERRORMSG_PROXY_REQUESTFAILED));
				break;
			case PROXYERROR_AUTHTYPEUNKNOWN:
				AddLogLine(false, RGB_LOG_ERROR + GetResString(IDS_ERRORMSG_PROXY_AUTHTYPEUNKNOWN));
				break;
			case PROXYERROR_AUTHFAILED:
				AddLogLine(false, RGB_LOG_ERROR + GetResString(IDS_ERRORMSG_PROXY_AUTHFAILED));
				break;
			case PROXYERROR_AUTHNOLOGON:
				AddLogLine(false, RGB_LOG_ERROR + GetResString(IDS_ERRORMSG_PROXY_AUTHNOLOGON));
				break;
			case PROXYERROR_CANTRESOLVEHOST:
				AddLogLine(false, RGB_LOG_ERROR + GetResString(IDS_ERRORMSG_PROXY_CANTRESOLVEHOST));
				break;
			default:
				AddLogLine(false, RGB_LOG_ERROR + _T("Unknown proxy error"));
			}
		}
	}
	EMULE_CATCH
	return 1;
}

