//this file is part of eMule
//Copyright (C)2002 Merkur ( merkur-@users.sourceforge.net / http://www.emule-project.net )
//
//This program is free software; you can redistribute it and/or
//modify it under the terms of the GNU General Public License
//as published by the Free Software Foundation; either
//version 2 of the License, or (at your option) any later version.
//
//This program is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.
//
//You should have received a copy of the GNU General Public License
//along with this program; if not, write to the Free Software
//Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

#include "StdAfx.h"
#include "zlib/zlib.h"
#include "UpDownClient.h"
#include "opcodes.h"
#include "packets.h"
#include "emule.h"
#include "UploadQueue.h"
#include "otherstructs.h"
#include "otherfunctions.h"
#include "math.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#define new DEBUG_NEW
#endif

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CUpDownClient::DrawUpStatusBar(CDC* dc, RECT* rect, bool bFlat)
{
	EMULE_TRY

	const COLORREF	crSentData = RGB(222, 160, 0);
	const COLORREF	crSendingPart = RGB(255, 208, 0);

//	Find the most used color
	COLORREF	crMostUsed;
	COLORREF	crLessUsed;
	byte		byteLessUsed;

	if ((GetAvailUpPartCount() * 2) < m_uUpPartCount)
	{
		crMostUsed = (bFlat) ? RGB(224, 224, 224) : RGB(240, 240, 240);
		crLessUsed = (bFlat) ? RGB(0, 150, 0) : RGB(0, 192, 0);
		byteLessUsed = 1;
	}
	else
	{
		crMostUsed = (bFlat) ? RGB(0, 150, 0) : RGB(0, 192, 0);
		crLessUsed = (bFlat) ? RGB(224, 224, 224) : RGB(240, 240, 240);
		byteLessUsed = 0;
	}

	CBarShader statusBar(rect->bottom - rect->top, rect->right - rect->left);

//	Get and set the file's size
	CKnownFile	   *pUploadedFile = g_eMuleApp.m_pSharedFilesList->GetFileByID(m_reqFileHash);
	uint32			dwFileSize;

	if (pUploadedFile != NULL)
	{
		dwFileSize = pUploadedFile->GetFileSize();
	}
	else
	{
		dwFileSize = m_uUpPartCount * PARTSIZE;
	}
	statusBar.SetFileSize(dwFileSize);

//	Fill the bar with the most used color
	statusBar.Fill(crMostUsed);

	uint32	dwUploadingPart = GetCurrentlyUploadingPart();
	uint32	dwLastPart = m_uUpPartCount - 1;
	uint32	dwStartPos;
	uint32	dwEndPos;

//	Draw the current uploading part if needed
	if (dwUploadingPart != 0xFFFF)
	{
		dwStartPos = PARTSIZE * dwUploadingPart;
		dwEndPos = (dwUploadingPart != dwLastPart) ? PARTSIZE * (dwUploadingPart + 1) : dwFileSize;
		statusBar.FillRange(dwStartPos, dwEndPos, crSendingPart);

		if (!m_doneBlocksList.IsEmpty())
		{
			Requested_Block_Struct	   *pDoneBlock;

			for (POSITION pos = m_doneBlocksList.GetHeadPosition(); pos != NULL; )
			{
				pDoneBlock = m_doneBlocksList.GetNext(pos);
				if (dwUploadingPart == (pDoneBlock->m_dwStartOffset / PARTSIZE))
				{
					statusBar.FillRange(pDoneBlock->m_dwStartOffset, pDoneBlock->m_dwEndOffset, crSentData);
				}
			}
		}
	}

//	Draw the parts with the less used color if needed
	if (m_pbyteUpPartStatuses != NULL && GetAvailUpPartCount() > 0 && (dwUploadingPart == 0xFFFF || GetAvailUpPartCount() < dwLastPart))
	{
		dwStartPos = 0;
		dwEndPos = 0;
		for (uint32 i = 0; i < m_uUpPartCount; i++)
		{
			if (m_pbyteUpPartStatuses[i] == byteLessUsed && i != dwUploadingPart)
			{
				if (dwEndPos != PARTSIZE * i)
				{
					if (dwStartPos != dwEndPos)
					{
						statusBar.FillRange(dwStartPos, dwEndPos, crLessUsed);
					}
					dwStartPos = PARTSIZE * i;
				}
				dwEndPos = PARTSIZE * (i + 1);
			}
		}
		if (dwEndPos > dwFileSize)
		{
			dwEndPos = dwFileSize;
		}
		statusBar.FillRange(dwStartPos, dwEndPos, crLessUsed);
	}

	statusBar.Draw(dc, rect->left, rect->top, bFlat);

	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
bool CUpDownClient::IsCommunity()
{
	EMULE_TRY

	if (!m_bCommunityIsCached)
	{
		CString username(m_strUserName);
		username.MakeLower();
		CString community(g_eMuleApp.m_pGlobPrefs->CommunityString());
		community.MakeLower();
		CString communityok;

		community.Remove( '[' );
		community.Remove(']');
		communityok.Format("[%s]", community);
		m_bIsCommunity = (username.Find(communityok) >= 0);
		m_bCommunityIsCached = true;
	}
	return m_bIsCommunity;

	EMULE_CATCH

	return false;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CUpDownClient::ResetValueCache()
{
	EMULE_TRY

	m_bScoreBaseIsCached = false;
	m_bCommunityIsCached = false;

	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
uint32 CUpDownClient::GetScore(bool sysvalue, bool isdownloading, bool onlybasevalue)
{
//	Spam control
	static DWORD	dwLastMsgTime = 0;

	EMULE_TRY

	if (m_strUserName.IsEmpty())
		return 0;

	if (m_bBanned)
	{
		return 0;
	}
	else
	{
		if ( IsStolenName(m_strUserName) || IsLeecherType(m_strModVersion) ||
			 (GetModVersion() == _T(PLUS_VERSION_STR) && GetMuleVersion() != CURRENT_VERSION_SHORT ) )
		{
		// Spam control
			if (::GetTickCount() > dwLastMsgTime + (60 * 1000))
			{
				if (!m_strModVersion.IsEmpty())
					AddDebugLogLine(false, RGB_LOG_DIMMED_TXT _T("Anti-leechermods: Client '%s' (%s) using the mod %s has been banned"), GetUserName(), GetClientNameAndVersionString(this), GetModVersion());
				else
					AddDebugLogLine(false, RGB_LOG_DIMMED_TXT _T("Anti-leechermods: Client '%s' (%s) has been banned"), GetUserName(), GetClientNameAndVersionString(this));

				Ban(_T("AUTOMATED MESSAGE: You're being banned because your client is leeching!"), GetResString(IDS_BAN_LEECHERADV));

				dwLastMsgTime = ::GetTickCount();
			}
			return 0;
		}
	}

#ifdef OLD_SOCKETS_ENABLED
	if (!isdownloading && !onlybasevalue)
	{
		if (HasLowID() && !(m_pRequestSocket && m_pRequestSocket->IsConnected()))
		{
			if (!g_eMuleApp.m_pServerConnect->IsConnected() || g_eMuleApp.m_pServerConnect->IsLowID())
				return 0;
		}
	}
#endif //OLD_SOCKETS_ENABLED

	CKnownFile* req_file = g_eMuleApp.m_pSharedFilesList->GetFileByID(m_reqFileHash);

	if (req_file == NULL)
		return 0;

// 	i_a: m_reqFileHash is an array so the if() was superfluous
	ASSERT(m_reqFileHash != NULL);

	double	dFilePriority;

	switch (req_file->GetPriority())
	{
		case PR_RELEASE:
			if (req_file->IsPartFile())
				dFilePriority = 5.0;
			else
				dFilePriority = 10.0;
			break;

		case PR_HIGH:
			dFilePriority = 2.0;
			break;

		case PR_LOW:
			dFilePriority = 0.5;
			break;

		case PR_VERYLOW:
			dFilePriority = 0.2;
			break;

		case PR_NORMAL:
		default:
			dFilePriority = 1.0;	// standard
			break;
	}

//	Calculate score, based on waiting time and other factors
	double		dblBaseValue;
	if (onlybasevalue)
	{
		dblBaseValue = 100.0f;
	}
	else if (!isdownloading)
	{
		dblBaseValue = static_cast<double>(GetTickCount() - m_dwWaitTime) / 1000.0;
	}
	else
	{
	//	We don't want one client to download forever
	//	the first 15 min download time counts as 15 min waitingtime and you get a 15 min bonus while you are in the first 15 min :)
	//	(to avoid 20 sec downloads) after this the score won't raise anymore
		dblBaseValue = static_cast<double>(m_dwUploadTime - m_dwWaitTime);
		ASSERT (m_dwUploadTime - m_dwWaitTime >= 0); //oct 28, 02: changed this from "> 0" to ">= 0"
		dblBaseValue += static_cast<double>((GetTickCount() - m_dwUploadTime > 900000) ? 900000.0 : 1800000.0);
		dblBaseValue /= 1000.0;
	}

//	Client is on friendlist
	if (IsFriend())
		dblBaseValue *= 2.0;

//	Client is a community user
	if (g_eMuleApp.m_pGlobPrefs->CommunityEnabled() && IsCommunity())
		dblBaseValue *= 2.0;

//	Client has credits (modifier between 1.0 and 10.0)
	if (m_pCredits != NULL)
	{
		dblBaseValue *= m_pCredits->GetScoreRatio(this->GetIP());
	}

//	Little penalty for _very_ old clients that are harmful for servers and network (from official)
	if((IsEmuleClient() && m_byteEmuleVersion <= 0x19) || this->GetClientSoft() == SO_SHAREAZA)
	{
		dblBaseValue *= 0.5f;
	}

//	Small and/or Rare file push
	dblBaseValue *= GetSmallFilePushRatio();
	dblBaseValue *= GetRareFilePushRatio();

	dblBaseValue *= dFilePriority;

	return static_cast<uint32>(dblBaseValue);

	EMULE_CATCH

	return 1;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
double CUpDownClient::GetSmallFilePushRatio()
{
	EMULE_TRY

	CKnownFile * srcfile = g_eMuleApp.m_pSharedFilesList->GetFileByID((uchar*)m_reqFileHash);

	if (srcfile == NULL)
	{
		return 1.0;
	}

	long lFileSize = srcfile->GetFileSize();

	if (lFileSize > 0.0 && lFileSize < 9961472)
	{
		double		dUploadModifier = 9961472.0 / lFileSize;

		if (dUploadModifier > 100.0)
		{
			dUploadModifier = 100.0;
		}

		return dUploadModifier;
	}

	EMULE_CATCH

	return 1.0;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
double CUpDownClient::GetRareFilePushRatio()
{
	EMULE_TRY

	CKnownFile * srcfile = g_eMuleApp.m_pSharedFilesList->GetFileByID((uchar*)m_reqFileHash);

	if (srcfile == NULL)
		return 1.0;

//	Keep the FileRatio
	double ratio = 0 + srcfile->GetFileRatio();

	return(ratio < 1.0 ? 1.0 : ((ratio > 100.0) ? 100.0 : ratio));

	EMULE_CATCH

	return 1.0;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//	Checks if it is next requested block from another chunk of the actual file or from another file
//	[Returns]
//	true : Next requested block is from another different chunk or file than last downloaded block
//	false: Next requested block is from same chunk that last downloaded block
bool CUpDownClient::IsDifferentPartBlock(bool silent)
{
	bool	different_part = false;

	try
	{
	//	Check if we have good lists and proceed to check for different chunks
		if (!m_blockRequestsQueue.IsEmpty() && !m_doneBlocksList.IsEmpty())
		{
		//	Get last block and next pending
			Requested_Block_Struct	*last_done_block = m_doneBlocksList.GetHead();
			Requested_Block_Struct	*next_requested_block = m_blockRequestsQueue.GetHead();

		//	Calculate corresponding parts to blocks
			uint32	last_done_part = last_done_block->m_dwStartOffset / PARTSIZE;
			uint32	next_requested_part = next_requested_block->m_dwStartOffset / PARTSIZE;

		//	Test is we are asking same file and same part
			if (last_done_part != next_requested_part || md4cmp(last_done_block->m_fileHash, next_requested_block->m_fileHash) != 0)
			{
				different_part = true;
			}
		}
	}
	catch (...)
	{
		different_part = true;
	}

	if (different_part)
	{
	//	obaldin: this is for old clients, that rotate the chunk, so that they'll be able to download at least _something_
	//	But probably even after the majority switches we'd better check for 0 here
		if (GetSessionUp() < 1024 * 1024)
		{
			different_part = false;

			if (!silent)
			{
				AddDebugLogLine(false, _T("Client '%s' (%s) continues downloading (should be back on queue) because of a low transferred size"), GetUserName(), GetClientNameAndVersionString(this));
			}
		}
	}

	return different_part;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
bool CUpDownClient::IsForbiddenByJumpstart()
{
	EMULE_TRY

	Requested_Block_Struct * next_requested_block;
	uint32 next_requested_part = 0xffffffff;

	if (m_blockRequestsQueue.IsEmpty())
	{
		return false;
	}
	else
	{
		next_requested_block = (Requested_Block_Struct*)m_blockRequestsQueue.GetHead();
		next_requested_part = next_requested_block->m_dwStartOffset / PARTSIZE;
		CKnownFile* kf = g_eMuleApp.m_pSharedFilesList->GetFileByID(next_requested_block->m_fileHash);

		if (kf != NULL && !kf->AllowChunkForClient(next_requested_part, this))
		{
			return true;
		}
	}

	EMULE_CATCH

	return false;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
class CSyncHelper
{
public:

	CSyncHelper()
	{
		m_pObject = NULL;
	}

	~CSyncHelper()
	{
		if (m_pObject)
			m_pObject->Unlock();
	}

	CSyncObject* m_pObject;
};
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//	Check if we should kick this client (time critical !!!)
//	- Always kick if transfer is longer than 80 minutes
//	- Never kick friends before a chunck was transferred
//	- Never allow more than the size of a chunk (+500 KBytes for small resends of invalid data) to be transferred
//	- We never kick a lancast user
//	- Full chunk transfers are being enforced, is not more an option
//
bool CUpDownClient::CreateNextBlockPackage()
{
	EMULE_TRY

	bool bAllowedMore = IsOnLAN();
	bool bDifferentPart = IsDifferentPartBlock(true);
	bool bJSIsBlocked = IsForbiddenByJumpstart();
	bool bShouldContinue = ((GetUpStartTimeDelay() >= 80 * 60 * 1000) || bDifferentPart);

//	If user is from Internet & Full chunk (or timeout)
	if (!bAllowedMore && bShouldContinue)
	{
		if ( IsFriend() && GetSessionUp() < (PARTSIZE + EMBLOCKSIZE * 4)
			 && !g_eMuleApp.m_pUploadQueue->CheckForTimeOver(this) && !bJSIsBlocked )
		{
			bAllowedMore = true;
			AddDebugLogLine(false, _T("Client '%s' (%s) downloaded %u KB so far. Allowed to continue downloading (should go back on queue) because it's a friend"), GetUserName(), GetClientNameAndVersionString(this), this->GetSessionUp() / 1024);
		}
		else
		{
			SetWaitStartTime();

			if (bDifferentPart)
			{
				g_eMuleApp.m_pUploadQueue->RemoveFromUploadQueue(this, ETS_BLOCKED_CHUNK);
				AddDebugLogLine(false, _T("Client '%s' (%s) back on queue. Asked for a new chunk/file, after receiving %u KB"), GetUserName(), GetClientNameAndVersionString(this), GetSessionUp() / 1024);
			}
			else if (bJSIsBlocked)
			{
				g_eMuleApp.m_pUploadQueue->RemoveFromUploadQueue(this, ETS_BLOCKED_CHUNK);
				AddDebugLogLine(false, _T("Client '%s' (%s) back on queue. Asked for a chunk forbidden by Jumpstart, after receiving %u KB"), GetUserName(), GetClientNameAndVersionString(this), GetSessionUp() / 1024);
			}
			else
			{
				g_eMuleApp.m_pUploadQueue->RemoveFromUploadQueue(this, ETS_TIMEOUT);
				AddDebugLogLine(false, _T("Client '%s' (%s) back on queue. Upload timeout, after receiving %u KB"), GetUserName(), GetClientNameAndVersionString(this), GetSessionUp() / 1024);
			}

			g_eMuleApp.m_pUploadQueue->AddClientToQueue(this, true);

			return false;
		}
	}

	if (m_blockRequestsQueue.IsEmpty())
	{
	//	There are no new blocks requested, we will happily transfer more if the client request it
		return false;
	}

	CFile			file;
	CKnownFile*		srcfile;
	byte*			filedata = 0;

	try
	{
	//	Repeat as long as next requested block is in the same chunk as previous blocks and
	//	as long as there is a next requested block. If a friend or lancast user, keep going!
		while ( !m_blockRequestsQueue.IsEmpty() && (bAllowedMore || !IsDifferentPartBlock()) &&
				!IsForbiddenByJumpstart() )
		{
			Requested_Block_Struct * currentblock = m_blockRequestsQueue.GetHead();
			srcfile = g_eMuleApp.m_pSharedFilesList->GetFileByID(currentblock->m_fileHash);

			if (srcfile == NULL)
			{
				CString srtError;

				g_eMuleApp.m_pUploadQueue->RemoveFromUploadQueue(this, ETS_FILE_ERROR);
				srtError.Format(_T("requested file not found, hash: %s"), HashToString(currentblock->m_fileHash));
				throw srtError;
			}

		//	Update the file hash to the requested file
			md4cpy(m_reqFileHash, currentblock->m_fileHash);

			if (!srcfile->GetJumpstartEnabled())
			{
			//	Accessing a blocked part ?
				byte part_status = srcfile->GetPartStatus(currentblock->m_dwStartOffset / PARTSIZE);
				CString sMessage;
				switch (part_status)
				{
					case PR_PART_OFF:
						AddLogLine( false, IDS_PARTPRIO_DENIED, currentblock->m_dwStartOffset / PARTSIZE,
									srcfile->GetFileName() );
						sMessage.Format( GetResString(IDS_PARTPRIO_DENIED), currentblock->m_dwStartOffset / PARTSIZE,
										 srcfile->GetFileName() );
						g_eMuleApp.m_pUploadQueue->RemoveFromUploadQueue(this, ETS_BLOCKED_CHUNK);
						throw sMessage;
						break;
					case PR_PART_HIDDEN:
						AddLogLine( false, IDS_PARTPRIO_HIDDEN, GetUserName(), srcfile->GetFileName(),
									currentblock->m_dwStartOffset / PARTSIZE );
				}
			}

			uint32 togo;
			if (bAllowedMore)
			{
			//	LAN clients are not limited by chunks boundaries.
			//	Is this a wraparound request?
				if (currentblock->m_dwStartOffset > currentblock->m_dwEndOffset - 1)
				{
				//	If m_dwStartOffset is later in the file than m_dwEndOffset, it means
				//	that the request is a wraparound request. It reads to the end of
				//	the file, then it wraps around and continues in the beginning of the file.

				//	Calculate how much the client is requesting, knowing that this is a wraparound request
					togo = currentblock->m_dwEndOffset + (srcfile->GetFileSize() - currentblock->m_dwStartOffset);
				}
				else
				{
				//	not a wraparound request. This makes it easier to calculate how many bytes is
				//	being requested.
					togo = currentblock->m_dwEndOffset - currentblock->m_dwStartOffset;
				}
			}
			else
			{
			//	Everyone else are limited to a single chunk
			//	Check that m_dwStartOffset and m_dwEndOffset is in the same chunk
				if (currentblock->m_dwStartOffset / PARTSIZE != (currentblock->m_dwEndOffset - 1) / PARTSIZE)
				{
				//	The m_dwEndOffset goes into the next chunk.
				//	Set it to the end of the chunk that m_dwStartOffset is in.
					currentblock->m_dwEndOffset = ((uint32)(currentblock->m_dwStartOffset / PARTSIZE) + 1) * PARTSIZE;
				}

			//	Make sure we don't pass end of file
				if (currentblock->m_dwStartOffset > currentblock->m_dwEndOffset)
				{
					currentblock->m_dwEndOffset = srcfile->GetFileSize();
				}

			//	This can't be a wrapped around request, since it has been limited to a single chunk.
				togo = currentblock->m_dwEndOffset - currentblock->m_dwStartOffset;
			}

			if (srcfile->IsPartFile())
			{
				uint32 parts_start = ((currentblock->m_dwStartOffset) / PARTSIZE) * PARTSIZE;
				uint32 parts_end = ((currentblock->m_dwEndOffset - 1) / PARTSIZE + 1) * PARTSIZE - 1;
				if (!(reinterpret_cast<CPartFile*>(srcfile))->IsComplete(parts_start, parts_end))
				{
					g_eMuleApp.m_pUploadQueue->RemoveFromUploadQueue(this, ETS_BLOCKED_CHUNK);
					throw CString(_T("asked for incomplete block"));
				}
			}

		//	Create a buffer for file data before we start to work with the file
			filedata = new byte[togo + 500];

			int	iRc = srcfile->ReadFileForUpload(currentblock->m_dwStartOffset, togo, filedata);

			if (iRc < 0)
			{
			//	Error occurred
				g_eMuleApp.m_pUploadQueue->RemoveFromUploadQueue(this, ETS_FILE_ERROR);
				throw CString(_T("failed to open requested file"));
			}

			bool	bFromPF = (iRc) ? true : false;

			md4cpy(m_reqFileHash, currentblock->m_fileHash);

			bool bCompress = false;

			if (m_byteDataCompVer == 1)
			{
				bCompress = true;

				CString	strExt = srcfile->GetFileName();

				int pos = strExt.ReverseFind('.');
				if (pos > 0)
				{
					static const TCHAR s_apcExt[][4] =
					{
						_T("zip"), _T("rar"), _T("ace"), _T("arj"),
						_T("cab"), _T("tgz"), _T("gz"), _T("bz2")
					};

				//	Take lowercase extension without point
					strExt = strExt.Mid(pos + 1);
					strExt.MakeLower();

				//	Don't try to compress already compressed files
					for (unsigned ui = 0; ui < ELEMENT_COUNT(s_apcExt); ui++)
					{
						if (strExt.Compare(reinterpret_cast<const TCHAR*>(&s_apcExt[ui])) == 0)
						{
							bCompress = false;
							break;
						}
					}
				}
			}
			if (bCompress)
			{
				CreatePackedPackets(filedata, togo, currentblock, bFromPF);
			}
			else
			{
				CreateStandardPackets(filedata, togo, currentblock, bFromPF);
			}

		//	File statistic
			srcfile->statistic.AddTransferred(togo);

			srcfile->statistic.AddTraffic(currentblock->m_dwStartOffset, togo);

			if (srcfile->GetJumpstartEnabled())
			{
				srcfile->AddSentBlock(this, currentblock->m_dwStartOffset, togo);
			}

			m_doneBlocksList.AddHead(m_blockRequestsQueue.RemoveHead());
			delete[] filedata;
			filedata = 0;
		}
	}
//	Check for exception that will be throw by "new"
	catch (CMemoryException * error)
	{
		AddDebugLogLine(false, RGB_LOG_ERROR + _T("Failed to create upload package for %s - %s"), GetUserName(), GetErrorMessage(error));
		g_eMuleApp.m_pUploadQueue->RemoveFromUploadQueue(this, ETS_FILE_ERROR);
		error->Delete();
		return false;
	}
	catch (CFileException * error)
	{
		OUTPUT_DEBUG_TRACE();
		AddDebugLogLine(false, RGB_LOG_ERROR + _T("Failed to create upload package for %s - %s"), GetUserName(), GetErrorMessage(error));
		g_eMuleApp.m_pUploadQueue->RemoveFromUploadQueue(this, ETS_FILE_ERROR);
		if (filedata)
			delete filedata;
		error->Delete();
		return false;
	}
	catch (CString & error)
	{
		OUTPUT_DEBUG_TRACE();
		AddDebugLogLine(false, _T("Client '%s' (%s) caused an error while creating package (%s), disconnecting client!"), GetUserName(), GetClientNameAndVersionString(this), error.GetBuffer());
		SetUploadFileID(NULL);
		if (filedata)
			delete filedata;
		return false;
	}

	return true;

	EMULE_CATCH

	return false;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CUpDownClient::CheckForUpPartsInfo()
{
	EMULE_TRY
	if (m_uPartCount == 0)
	{
		return;
	}
	if (!m_pReqPartFile)
	{
		return;
	}
	if (md4cmp(m_reqFileHash, m_pReqPartFile->GetFileHash()))
	{
		return;
	}
	if (!m_pbytePartStatuses)
	{
		return;
	}
	if (m_pbyteUpPartStatuses)
	{
		delete[] m_pbyteUpPartStatuses;
		m_pbyteUpPartStatuses = NULL;
	}

	m_uUpPartCount = m_uPartCount;
	m_uAvailUpPartCount = 0;
	m_pbyteUpPartStatuses = new byte[m_uUpPartCount];

	for (int i = 0; i < m_uUpPartCount; i++)
	{
		m_pbyteUpPartStatuses[i] = m_pbytePartStatuses[i];

		if (m_pbyteUpPartStatuses[i])
		{
			m_uAvailUpPartCount++;
		}
	}
	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CUpDownClient::ProcessUpFileStatus(char* packet, uint32 size)
{
	EMULE_TRY

	if (m_pbyteUpPartStatuses)
	{
		delete[] m_pbyteUpPartStatuses;
		m_pbyteUpPartStatuses = NULL;
	}
	m_uUpPartCount = 0;
	m_uAvailUpPartCount = 0;
	m_uUpCompleteSourcesCount = 0;
	if (size == 16)
	{
		CheckForUpPartsInfo();
		return;
	}
	CSafeMemFile data((BYTE*)packet, size);
	uchar cfilehash[16];
	data.Read(cfilehash, 16);
	CKnownFile* tempreqfile = g_eMuleApp.m_pSharedFilesList->GetFileByID(cfilehash);
	uint16 uED2KUpPartCount;

	data.Read(&uED2KUpPartCount, 2);
	m_uUpPartCount = tempreqfile->GetPartCount();

	if (uED2KUpPartCount == 0)
	{
		m_pbyteUpPartStatuses = new byte[m_uUpPartCount];
		memzero(m_pbyteUpPartStatuses, m_uUpPartCount);
	}
	else
	{
		if (tempreqfile->GetED2KPartCount() != uED2KUpPartCount)
		{
			m_uUpPartCount = 0;
			return;
		}

		m_pbyteUpPartStatuses = new byte[m_uUpPartCount];
		uint16 done = 0;

		while (done != m_uUpPartCount)
		{
			byte toread;
			data.Read(&toread, 1);

			for (sint32 i = 0;i != 8;i++)
			{
				m_pbyteUpPartStatuses[done] = ((toread >> i) & 1) ? 1 : 0;

				if (m_pbyteUpPartStatuses[done])
				{
					m_uAvailUpPartCount++;
				}
				done++;
				if (done == m_uUpPartCount)
				{
					break;
				}
			}
		}

		if ((GetExtendedRequestsVersion() > 1) && (size >= 20 + (((uint32)m_uUpPartCount + 7) >> 3)))
		{
			uint16 nCompleteCountLast = GetUpCompleteSourcesCount();
			uint16 nCompleteCountNew;
			data.Read(&nCompleteCountNew, 2);
			SetUpCompleteSourcesCount(nCompleteCountNew);
			if (nCompleteCountLast != nCompleteCountNew)
				tempreqfile->CalculateCompleteSources();
		}
	}
	g_eMuleApp.m_pdlgEmule->m_wndTransfer.m_ctlQueueList.UpdateClient(this);

	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CUpDownClient::CreateStandardPackets(byte* pbyteFileData, uint32 togo, Requested_Block_Struct* currentblock, bool bFromPF)
{
	EMULE_TRY

	uint32			nPacketSize;
	CMemFile		memfile(pbyteFileData, togo);

	if (togo > 10240)
	{
		nPacketSize = togo / (togo / 10240L);
	}
	else
	{
		nPacketSize = togo;
	}

	while (togo != 0)
	{
		if (togo < nPacketSize * 2)
		{
			nPacketSize = togo;
		}

		togo -= nPacketSize;

		Packet		*packet = new Packet(OP_SENDINGPART, nPacketSize + 24, OP_EDONKEYPROT, bFromPF);

		md4cpy(&packet->m_pcBuffer[0], m_reqFileHash);

		uint32 startpos = (currentblock->m_dwEndOffset - togo) - nPacketSize;
		uint32 endpos = (currentblock->m_dwEndOffset - togo);

		memcpy2(&packet->m_pcBuffer[16], &startpos, 4);
		memcpy2(&packet->m_pcBuffer[20], &endpos, 4);

	//	Append the file data to the packet
		memfile.Read(&packet->m_pcBuffer[24], nPacketSize);

		SplitAndAddPacketsToSendQueue(packet, bFromPF);
	}

	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CUpDownClient::CreatePackedPackets(byte* pbyteFileData, uint32 togo, Requested_Block_Struct* currentblock, bool bFromPF)
{
	EMULE_TRY

	BYTE * output = new BYTE[togo + 300];
	uLongf newsize = togo + 300;
	uint16 result = compress2(output, &newsize, pbyteFileData, togo, 9);

	if (result != Z_OK || togo <= newsize)
	{
		delete[] output;
		CreateStandardPackets(pbyteFileData, togo, currentblock, bFromPF);
		return;
	}

	m_dwCompressionGain += (togo - newsize);
	m_dwUncompressed += togo;
	CMemFile memfile(output, newsize);
	togo = newsize;
	uint32 nPacketSize;

	if (togo > 10240)
	{
		nPacketSize = togo / (uint32)(togo / 10240);
	}
	else
	{
		nPacketSize = togo;
	}

	while (togo)
	{
		if (togo < nPacketSize * 2)
		{
			nPacketSize = togo;
		}
		togo -= nPacketSize;
		Packet* packet = new Packet(OP_COMPRESSEDPART, nPacketSize + 24, OP_EMULEPROT, bFromPF);
		md4cpy(&packet->m_pcBuffer[0], m_reqFileHash);

		uint32 startpos = currentblock->m_dwStartOffset;

		memcpy2(&packet->m_pcBuffer[16], &startpos, 4);
		memcpy2(&packet->m_pcBuffer[20], &newsize, 4);
		memfile.Read(&packet->m_pcBuffer[24], nPacketSize);

		SplitAndAddPacketsToSendQueue(packet, bFromPF);
	}
	delete[] output;

	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// A packet should never be larger than MAXFRAGSIZE. This method splits up the
// given packets into several packets with the size MAXFRAGSIZE, if necessary.
void CUpDownClient::SplitAndAddPacketsToSendQueue(Packet* packet, bool bFromPF)
{
	EMULE_TRY

	if (packet->GetRealPacketSize() > 0 && packet->GetRealPacketSize() > MAXFRAGSIZE * 2)
	{
	//	Splitting packets
		uint32 nSize = packet->GetRealPacketSize();
		char* m_pcBuffer = packet->DetachPacket();
		delete packet;

		uint32 nPos = 0;
		bool bLast = true;
		while (nPos < nSize)
		{
			uint32 nNewSize = (nSize - nPos < MAXFRAGSIZE) ? nSize - nPos : MAXFRAGSIZE;
			char* pBuffer2 = new char[nNewSize];
			memcpy2(pBuffer2, m_pcBuffer + nPos, nNewSize);
			nPos += nNewSize;
			m_blockSendQueue.AddTail(new Packet(pBuffer2, nNewSize, (nNewSize < MAXFRAGSIZE), bFromPF));
			bLast = false;
		}
		delete[] m_pcBuffer;
	}
	else
	{
		m_blockSendQueue.AddTail(packet);
	}

	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CUpDownClient::SetUploadFileID(uchar *tempreqfileid)
{
	EMULE_TRY

	if (tempreqfileid)
	{
		md4cpy(m_reqFileHash, tempreqfileid);
	}
	else
	{
		md4clr(m_reqFileHash);
	}

	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CUpDownClient::AddReqBlock(Requested_Block_Struct* pNewReqBlock)
{
	EMULE_TRY

	for (POSITION pos = m_doneBlocksList.GetHeadPosition(); pos != NULL; )
	{
		Requested_Block_Struct* pFinishedBlock = static_cast<Requested_Block_Struct*>(m_doneBlocksList.GetNext(pos));

		if (pFinishedBlock && pNewReqBlock->m_dwStartOffset == pFinishedBlock->m_dwStartOffset
			&& pNewReqBlock->m_dwEndOffset == pFinishedBlock->m_dwEndOffset)
		{
			return;
		}
	}

	for (POSITION pos = m_blockRequestsQueue.GetHeadPosition(); pos != NULL; )
	{
		Requested_Block_Struct* pReqBlock = static_cast<Requested_Block_Struct*>(m_blockRequestsQueue.GetNext(pos));

		if (pReqBlock && pNewReqBlock->m_dwStartOffset == pReqBlock->m_dwStartOffset
			&& pNewReqBlock->m_dwEndOffset == pReqBlock->m_dwEndOffset)
		{
			return;
		}
	}

	Requested_Block_Struct* pNewBlock = new Requested_Block_Struct;

	memcpy2(pNewBlock, pNewReqBlock, sizeof(Requested_Block_Struct));
	m_blockRequestsQueue.AddTail(pNewBlock);

	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CUpDownClient::SetUpStartTime(uint32 dwTime)
{
	EMULE_TRY

	if (dwTime != 0)
	{
		m_dwUploadTime = dwTime;
	}
	else
	{
		m_dwUploadTime = ::GetTickCount();
	}

	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CUpDownClient::SetWaitStartTime(uint32 dwTime)
{
	EMULE_TRY

	if (dwTime != 0)
	{
		m_dwWaitTime = dwTime;
	}
	else
	{
		m_dwWaitTime = ::GetTickCount();
	}

	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
uint32 CUpDownClient::SendBlockData(uint32 nMaxAmmount, bool& wantsMoreBandwidth)
{
	EMULE_TRY

	DWORD curTick = GetTickCount();
//	This will be set to true if/when the slot is actually fed some data
	wantsMoreBandwidth = false;
	uint32 amounttransferred = 0;

//	Check for 'stuck' clients
	if ((curTick - GetLastGotULData() > 2 * 60 * 1000) && (GetUpStartTimeDelay() > 2 * 60 * 1000))
	{
	//	We were unable to transfer for 2 minutes, put the client back on queue
		AddDebugLogLine( false, _T("Client '%s' (%s) appears to be stuck, putting back on queue"),
						 GetUserName(), GetClientNameAndVersionString(this) );
		SetWaitStartTime();
		g_eMuleApp.m_pUploadQueue->RemoveFromUploadQueue(this, ETS_TIMEOUT);
		g_eMuleApp.m_pUploadQueue->AddClientToQueue(this, true);
		return 0;
	}

#ifdef OLD_SOCKETS_ENABLED
//	Only feed this slot new data if it has been able to empty
//	its buffer since we last offered data
	if (m_pRequestSocket && !m_pRequestSocket->IsBusy() && GetUploadState() == US_UPLOADING)
	{
		if (m_blockSendQueue.IsEmpty())
		{
		//	No valid blocks to transfer. Put the client back on queue
			if (!CreateNextBlockPackage())
			{
				return 0;
			}
		}

	//	Feed data to this slot, one block at a time, as long as there's enough bandwidth left
	//	for the entire next block. There must also be blocks to send (not empty request queue)
		while ( !m_blockSendQueue.IsEmpty() &&
				m_blockSendQueue.GetHead()->GetRealPacketSize() + amounttransferred <= nMaxAmmount )
		{
		//	Get next block and give it to the Request Socket for transfer
			Packet		*pPacket = m_blockSendQueue.RemoveHead();
			uint32		nBlockSize = pPacket->GetRealPacketSize();

		//	Extended statistics information based on which client software and which port we sent this data to.
		//	This also updates the grand total for sent bytes, etc. And where this data came from.
		//	We have to call function before packet will be sent, otherwise packet will be deleted.
			g_eMuleApp.m_pGlobPrefs->Add2SessionTransferData(GetClientSoft(), GetUserPort(), pPacket->IsFromPF(), TRANSFERTYPE_UPLOAD, nBlockSize, IsCommunity());
			m_pRequestSocket->SendPacket(pPacket, true, false);
			pPacket = NULL;

		//	We've transferred some data. Store this in relevant statistics variables
			m_dwTransferredUp += nBlockSize;
			amounttransferred += nBlockSize;
			m_pCredits->AddUploaded(nBlockSize, this->GetIP());

		//	If we're out of prepared blocks, create some more
			if (m_blockSendQueue.IsEmpty())
			{
				CreateNextBlockPackage();
			}
		}

	//	If this slot wasn't given enough data to empty its request queue, it wants more data.
		if (!m_blockSendQueue.IsEmpty())
		{
			wantsMoreBandwidth = true;
		}

	//	Remember when this connection was last fed some data, so we
	//	can make sure we give it data often enough to not time out.
		if (amounttransferred > 0)
		{
			SetLastGotULData();
		}
	}
#endif //OLD_SOCKETS_ENABLED

//	Store how much data we've transferred this round,
//	to be able to calculate average speed later
	m_averageUDRList.push_front(m_dwTransferredUp);
	m_averageULTickList.push_front(curTick);

//	Don't store too much statistics
	while (m_averageULTickList.size() > 500)
	{
		m_averageUDRList.pop_back();
		m_averageULTickList.pop_back();
	}

//	Since performance is more importatnt than graphical output, this function was moved to the end
//	Check if it's time to update the display.
	if (curTick - m_dwLastRefreshedULDisplayTime > MINWAIT_BEFORE_ULDISPLAY_WINDOWUPDATE && m_averageUDRList.size() >= 2)
	{
	//	Calculate average data rate
	//	Present this slot's speed as 0 unless there's a good chunk of stored statistics for it.
	//	If there's to little data, the UL speed will be far off, so we might as well hide it in those cases.
		uint32 timePeriod = m_averageULTickList.front() - m_averageULTickList.back();
		if (timePeriod > 5000)
		{
			m_lUpDataRate = static_cast<sint32>((static_cast<double>(m_averageUDRList.front() - m_averageUDRList.back())) * 1000.0 / timePeriod);
		}
		else
		{
			m_lUpDataRate = -1;
		}
		g_eMuleApp.m_pdlgEmule->m_wndTransfer.m_ctlUploadList.UpdateClient(this);
		m_dwLastRefreshedULDisplayTime = curTick + (uint32)(rand() / (RAND_MAX / 1000));
	}

	return amounttransferred;

	EMULE_CATCH

	return 0;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//	FlushSendBlocks() sends the rest of the last block queued for upload to this client if it has been split
//	across multiple packets. Call this when you stop uploading or the request socket is not able to send.
void CUpDownClient::FlushSendBlocks()
{
	EMULE_TRY

	bool	bBreak = false;

#ifdef OLD_SOCKETS_ENABLED
	while ( !m_blockSendQueue.IsEmpty() && m_blockSendQueue.GetHead()->IsSplit() &&
			m_pRequestSocket != NULL && m_pRequestSocket->IsConnected() && !bBreak )
	{
		Packet	*pPacket = m_blockSendQueue.RemoveHead();
		bool	bBreak = pPacket->IsLastSplit();

		g_eMuleApp.m_pUploadQueue->AddUpDataOverheadOther(pPacket->m_dwSize);
		m_pRequestSocket->SendPacket(pPacket, true, false);
	}
#endif //OLD_SOCKETS_ENABLED

	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CUpDownClient::SendHashsetPacket(char* fileHash)
{
	EMULE_TRY

	CKnownFile		*pKnownFile = g_eMuleApp.m_pSharedFilesList->GetFileByID(reinterpret_cast<BYTE*>(fileHash));

	if (pKnownFile == NULL)
	{
		AddDebugLogLine(false, RGB_LOG_ERROR + _T("Client '%s' (%s): requested file not found (hash: %s)"), GetUserName(), GetClientNameAndVersionString(this), HashToString(reinterpret_cast<BYTE*>(fileHash)));
	//	cause OP_HASHSETREQUEST can be abused for file-scanning
		UpdateFailedFileRequests();
		return;
	}

	CSafeMemFile		pbyteFileData(1024);

	pbyteFileData.Write(pKnownFile->GetFileHash(), 16);

	uint16		uPartCount = pKnownFile->GetHashCount();

	pbyteFileData.Write(&uPartCount, 2);

	for (int i = 0; i < uPartCount; i++)
	{
		pbyteFileData.Write(pKnownFile->GetPartHash(i), 16);
	}

	Packet		*packet = new Packet(&pbyteFileData);

	packet->m_eOpcode = OP_HASHSETANSWER;
	g_eMuleApp.m_pUploadQueue->AddUpDataOverheadFileRequest(packet->m_dwSize);
#ifdef OLD_SOCKETS_ENABLED

	m_pRequestSocket->SendPacket(packet, true, true);
#endif //OLD_SOCKETS_ENABLED

	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
double CUpDownClient::GetClientSpeed()
{
	EMULE_TRY
//	Calculates the session speed of the client. Providing the other Clients with this info
//	gives then the possibility to select the sources more wisely.
	if (g_eMuleApp.stat_transferStarttime == 0)
	{
		return 0.0;
	}

	DWORD running = (GetTickCount() - g_eMuleApp.stat_transferStarttime) / 1000;

	if (running < 5)
	{
		return 0.0;
	}

	return static_cast<double>(g_eMuleApp.stat_sessionSentBytes / 1024.0) / running;

	EMULE_CATCH

	return 0.0;
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// FlushSendBlocks() was removed, since the client on the other side ignores all packets,
// that will be coming after an OP_CANCEL... what actually means traffic waste.
void CUpDownClient::ClearUploadBlockRequests()
{
	EMULE_TRY

	while (!m_blockRequestsQueue.IsEmpty())
	{
		delete m_blockRequestsQueue.RemoveHead();
	}

	while (!m_blockSendQueue.IsEmpty())
	{
		delete m_blockSendQueue.RemoveHead();
	}

	while (!m_doneBlocksList.IsEmpty())
	{
		delete m_doneBlocksList.RemoveHead();
	}

	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CUpDownClient::SendRankingInfo()
{
	EMULE_TRY

#ifdef OLD_SOCKETS_ENABLED
	if (m_pRequestSocket == NULL)
	{
		return;
	}
#endif //OLD_SOCKETS_ENABLED

	if (!ExtProtocolAvailable())
	{
		return;
	}
	uint16 nRank = g_eMuleApp.m_pUploadQueue->GetWaitingPosition(this);

#ifndef DISABLE_EXTENDED_INFO_EXCHANGE
 #define DISABLE_EXTENDED_INFO_EXCHANGE
// The extended info exchange is currently disabled since it is possible
// to ask the client with several userhashes for the file and then use
// the one with the best UL/DL ratio
// It is planned to reactivate this feature, but with a protection against
// this userhash spoofing.
// At the last stage this faeture should make it possible to choose the
// sources more intelligently by taking UL/DL ratio, queuerank and
// speed of the other client into account.
#endif

#ifdef DISABLE_EXTENDED_INFO_EXCHANGE
	if (!nRank)
	{
		return;
	}
#else

	uint16 nBaseModifier = (uint16) GetScore(false, false, true);
	uint32 nScore = GetScore(false, false, false);

	byte nCommunity = -1;

	if (!g_eMuleApp.m_pGlobPrefs->CommunityEnabled())
	{
		nCommunity = 2;
	}
	else if (IsCommunity())
	{
		nCommunity = 1;
	}

	double dblBaseValue = 1.0;

	if (m_pCredits != (CClientCredits*)NULL)
	{
		dblBaseValue *= m_pCredits->GetScoreRatio(this->GetIP());
	}

	dblBaseValue *= 10;
	uint16 nRatio = (uint16)dblBaseValue;
#endif

	uint16 nClientSpeed = (uint16)((GetClientSpeed() < 65535) ? GetClientSpeed() : 65535);

//	Providing the other clients with this info can not be used to leech
	Packet* packet = new Packet(OP_QUEUERANKING, 12, OP_EMULEPROT);
	memzero(packet->m_pcBuffer, 12);
	memcpy2(packet->m_pcBuffer + 0, &nRank, 2);
#ifndef DISABLE_EXTENDED_INFO_EXCHANGE

	memcpy2(packet->m_pcBuffer + 2, &nBaseModifier, 1);		// UINT8 is enough 0-100
	memcpy2(packet->m_pcBuffer + 3, &nRatio, 2);			// UINT16 needed saw about 32000 max (SF * UL/DL * RF)
	memcpy2(packet->m_pcBuffer + 5, &nScore, 4);			// UINT32 needed
	memcpy2(packet->m_pcBuffer + 9, &nCommunity, 1);		// 2-Bits needed
#endif

	memcpy2(packet->m_pcBuffer + 10, &nClientSpeed, 2);		// UINT16 (I think not many have more than 65 MByte UL)

	g_eMuleApp.m_pUploadQueue->AddUpDataOverheadOther(packet->m_dwSize);
#ifdef OLD_SOCKETS_ENABLED

	m_pRequestSocket->SendPacket(packet, true, true);
#endif //OLD_SOCKETS_ENABLED

	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CUpDownClient::SendCommentInfo(CKnownFile *pKnownFile)
{
	EMULE_TRY

	if (!m_bCommentDirty || pKnownFile == NULL || !ExtProtocolAvailable() || m_byteAcceptCommentVer < 1)
	{
		return;
	}
	m_bCommentDirty = false;

	EnumPartFileRating rating = pKnownFile->GetFileRating();
	CString desc = pKnownFile->GetFileComment();
	if (pKnownFile->GetFileRating() == PF_RATING_NONE && desc.IsEmpty())
	{
		return;
	}

	CSafeMemFile		packetStream(256);

	packetStream.Write(&rating, sizeof(rating));		// <rating 1>

	int		length = desc.GetLength();

	if (length > 128)
	{
		length = 128;
	}
	packetStream.Write(&length, sizeof(length));		// <commentlen 4>

	if (length > 0)
	{
		packetStream.Write(desc.GetBuffer(), length);	// (<commentchars 1>)*commentlen
	}

	Packet		*packet = new Packet(&packetStream, OP_EMULEPROT);

	packet->m_eOpcode = OP_FILEDESC;
	g_eMuleApp.m_pUploadQueue->AddUpDataOverheadOther(packet->m_dwSize);
#ifdef OLD_SOCKETS_ENABLED

	m_pRequestSocket->SendPacket(packet, true);
#endif //OLD_SOCKETS_ENABLED

	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CUpDownClient::AddRequestCount(uchar* fileid)
{
	EMULE_TRY

	for (POSITION pos = m_requestedFilesList.GetHeadPosition(); pos != NULL; )
	{
		Requested_File_Struct	   *pReqFile = static_cast<Requested_File_Struct*>(m_requestedFilesList.GetNext(pos));

		if (!md4cmp(pReqFile->m_fileHash, fileid))
		{
			if (::GetTickCount() - pReqFile->m_dwLastAskedTime < MIN_REQUESTTIME)
			{
				if (GetDownloadState() != DS_DOWNLOADING)
				{
					pReqFile->m_byteNumBadRequests++;
				}
				if (pReqFile->m_byteNumBadRequests == BADCLIENTBAN)
				{
					Ban( _T("AUTOMATED MESSAGE: You're being banned because your client is behaving aggressively!"),
						 GetResString(IDS_BAN_AGGRESSIVECLI) );
				}
			}
			else
			{
				if (pReqFile->m_byteNumBadRequests)
				{
					pReqFile->m_byteNumBadRequests--;
				}
			}
			pReqFile->m_dwLastAskedTime = ::GetTickCount();

			return;
		}
	}
	Requested_File_Struct	   *pNewFile = new Requested_File_Struct;
	memzero(pNewFile, sizeof(Requested_File_Struct));
	md4cpy(pNewFile->m_fileHash, fileid);
	pNewFile->m_dwLastAskedTime = ::GetTickCount();
	m_requestedFilesList.AddHead(pNewFile);

	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CUpDownClient::UnBan()
{
	EMULE_TRY

	m_bBanned = false;
	m_dwBanTime = 0;

	SetWaitStartTime();
	g_eMuleApp.m_pUploadQueue->UpdateBanCount();
	g_eMuleApp.m_pdlgEmule->m_wndTransfer.UpdateUploadHeader();

	const uint64 qwIPPortID = GetUID();

	g_eMuleApp.m_pClientList->m_mapOffenseCounter.erase(qwIPPortID);
	g_eMuleApp.m_pClientList->m_mapHashBase.erase(qwIPPortID);
	m_bIsHashThief = false;

	for (POSITION pos = m_requestedFilesList.GetHeadPosition(); pos != NULL; )
	{
		Requested_File_Struct *pReqFile = static_cast<Requested_File_Struct*>(m_requestedFilesList.GetNext(pos));
		pReqFile->m_byteNumBadRequests = 0;
		pReqFile->m_dwLastAskedTime = 0;
	}
	g_eMuleApp.m_pdlgEmule->m_wndTransfer.m_ctlQueueList.UpdateClient(this);

	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CUpDownClient::Ban(LPCTSTR line, const CString& reason, ...)
{
	EMULE_TRY

	if (!g_eMuleApp.m_pGlobPrefs->BanEnabled())
	{
	//	If the ban function is disabled, no need to ban a client (default is enabled!)
		m_bBanned = false;
		m_dwBanTime = 0;
		return;
	}

	if (m_eUploadState == US_UPLOADING)
	{
	//	Kick already downloading client back to waitingqueue
		AddDebugLogLine(false, RGB_LOG_DIMMED + _T("Client '%s' (%s) banned and kicked out of uploadqueue %s"), GetUserName(), GetClientNameAndVersionString(this), reason);
		g_eMuleApp.m_pUploadQueue->RemoveFromUploadQueue(this, ETS_BAN, true);
		g_eMuleApp.m_pUploadQueue->AddClientToQueue(this, true);
	}

	if (m_bBanned)
	{
	//	No need to ban a client (and send a message) again except for extending the bantime
		m_dwBanTime = ::GetTickCount();
		return;
	}

	if (g_eMuleApp.m_pGlobPrefs->CommunityNoBanEnabled() && g_eMuleApp.m_pGlobPrefs->CommunityEnabled() && IsCommunity())
	{
		return;
	}
	m_bBanned = true;

	g_eMuleApp.m_pUploadQueue->UpdateBanCount();
	g_eMuleApp.m_pdlgEmule->m_wndTransfer.UpdateUploadHeader();

	m_dwBanTime = ::GetTickCount();

	if (g_eMuleApp.m_pGlobPrefs->IsBanMessageEnabled())
	{
		g_eMuleApp.m_pdlgEmule->m_wndChat.chatselector.StartSession(this);
		g_eMuleApp.m_pdlgEmule->m_wndChat.chatselector.SendMessage(line);
		g_eMuleApp.m_pdlgEmule->m_wndChat.chatselector.EndSession(this);
		if (!g_eMuleApp.m_pGlobPrefs->IsCMNotLog())
		{
			AddLogLine(false, RGB_LOG_DIMMED + GetResString(IDS_SENTBAN), GetUserName(), reason);
		}
	}
	else
	{
		if (!g_eMuleApp.m_pGlobPrefs->IsCMNotLog())
		{
			AddLogLine(false, RGB_LOG_DIMMED + GetResString(IDS_CLIENTBLOCKED), GetUserName(), reason);
		}
	}

	g_eMuleApp.m_pdlgEmule->m_wndTransfer.m_ctlQueueList.UpdateClient(this);

	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
CString CUpDownClient::GetDownloadStateAsString(void)
{
	return GetStatusDLQueueString((EnumDLQState)GetDownloadState());
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
uint32	CUpDownClient::GetRemoteScore()
{
	EMULE_TRY

	if (IsRemoteInfoAvaiable())
	{
		return m_dwRemoteScore;
	}

	EMULE_CATCH

//	We do not know for how long we are on the queue.
//	Furhtermore the we do not know the remote ratio.
	return 0;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
uint16	CUpDownClient::GetRemoteRatio()
{
	EMULE_TRY
	if (IsRemoteInfoAvaiable())
		return m_uRemoteRatio;

//	This calculation is not very accurate for several reasons:
//	- We do not know the file priority
//	- We do not save the client info if he has nothing uploaded to us
//	- Other factores like community etc

	double dblBaseValue = 100.0;

#ifdef OLD_SOCKETS_ENABLED

	if (g_eMuleApp.m_pServerConnect->IsLowID())
	{
		dblBaseValue *= 0.8;
	}
#endif //OLD_SOCKETS_ENABLED

	dblBaseValue *= GetRemoteBaseModifier();
//	Official client doesnt have SF/RF push
	if (this->GetPlusVersion() != 0)
	{
		dblBaseValue *= GetSmallFilePushRatio() ;
		dblBaseValue *= GetRareFilePushRatio() ;
	}
	return static_cast<uint32>(dblBaseValue);

	EMULE_CATCH

	return 100;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
double CUpDownClient::GetRemoteBaseModifier()
{
	EMULE_TRY

	if (IsRemoteInfoAvaiable())
	{
		return static_cast<double>(m_uRemoteBaseModifier) / 100.0;
	}
//	This calculation is not very accurate for some reasons...
//	Crashes can cause that either client can not write its client data.
//	This can cause differences of the client data at the two clients.
	if (m_pCredits == reinterpret_cast<CClientCredits*>(-1))
	{
		m_pCredits = NULL;
	}
	if (m_pCredits != NULL)
	{
		return m_pCredits->GetScoreRatio(this->GetIP(), true);
	}

	EMULE_CATCH

	return 1.0;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
uint32 CUpDownClient::GetCurrentlyUploadingPart()
{
	EMULE_TRY

	if (!m_doneBlocksList.IsEmpty())
	{
	//	Do we need to check upload state here?
		Requested_Block_Struct		*pFirstDoneBlock = m_doneBlocksList.GetHead();

	//	Calculate which part the block is in.
		uint32		dwPartNum = pFirstDoneBlock->m_dwStartOffset / PARTSIZE;

		return dwPartNum;
	}
	else
	{
		return 0xFFFF;
	}

	EMULE_CATCH

	return 0xFFFF;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CUpDownClient::SetUploadState(EnumULQState news)
{
	m_eUploadState = news;
	g_eMuleApp.m_pdlgEmule->m_wndTransfer.m_ctlClientList.UpdateClient(this);
}
